package biggests.tables;

import java.awt.*;
import java.awt.event.*;

import javax.swing.table.*;

import biggests.utils.*;
import smadeira.biclustering.*;
import smadeira.ontologizer.*;

/**
 * <p>Title: Table of colors with missing values</p>
 *
 * <p>Description:
 *                 <p>Defines a table of colors with the following data:<br/>
 *                 - header -> conditions' names (more widely, <code>String</code>s)<br/>
 *                 - first column -> genes' names (more widely, <code>String</code>s)<br/>
 *                 - all other columns' cells -> gene expression values
 *                 (more widely, <code>float</code> values)</p>
 *
 *                 <p>A table of colors has a vertical text header and each
 *                 of its cells is colored according to the float value it
 *                 holds.</p>
 *
 * <p> The cells which value is situated in the interval:<br/> <b>[minValue,
 * <code>meanValue</code>[</b> are colored green (different shades of green,
 * according to the exact value)<br/> <b>{<code>meanValue</code>}</b> are
 * colored black<br/> <b>]<code>meanValue</code>, maxValue]</b> are colored
 * red (different shades of red, according to the exact value)<br/> cells of
 * missing values are colored yellow </p>
 * <p>
 * This table also displays the terms annotated with each gene
 * (when a specific row of the table is clicked by the user).</p>
 *
 *                 <p>This kind of table implements sorting functionality
 *                 on the columns, i.e., allows the reordering of the rows,
 *                 sorting them by the values of the cells of a given column
 *                 in ascending or descending order.</p>
 *
 *                 <p>Specific for data with missing values, that is, some
 *                 erroneus values that are not part of the original data
 *                 and are marked by some float value. Missing values are
 *                 colored with a different color from that of the other
 *                 values.</p>
 * </p>
 *
 * <p>Copyright:   Copyright (C) 2007  Joana Gonçalves
 *                 This program is free software; you can redistribute
 *                 it and/or modify it under the terms of the GNU General
 *                 Public License as published by the Free Software
 *                 Foundation; either version 3 of the License, or
 *                 any later version.
 *
 *                 This program is distributed in the hope that it will
 *                 be useful, but WITHOUT ANY WARRANTY; without even the
 *                 implied warranty of MERCHANTABILITY or FITNESS FOR A
 *                 PARTICULAR PURPOSE.  See the
 *                 <a href="http://www.gnu.org/licenses/gpl.html">
 *                 GNU General Public License</a> for more details.
 * </p>
 *
 * @author Joana P. Gonçalves
 */
/* A colored table looks like this:<br/>
 *   ------------------------------------------------------------<br/>
 *   | Genes Names | C1 | C2 | C3 |   ......................    |<br/>
 *   |----------------------------------------------------------|<br/>
 *   | Gene1 | o | o | o | .................................... |<br/>
 *   | Gene2 | o | o | o | .................................... |<br/>
 *   | Gene3 | o | o | o | .................................... |<br/>
 *   | Gene4 | o | o | o | .................................... |br/>
 *   ------------------------------------------------------------<br/>
 *   where 'o' represents a colored cell. Conditions' names are written
 *   with vertical orientation.
 */
public class TableColorsWithMissingValues extends TableColors {
    /**
     * The value that marks the cells which "contain" missing values.
     */
    private float missingValue;

    /**
     * Creates a new colored table with data from the given expression matrix
     * with missing values.
     *
     * @param expressionMatrix the <code>AbstractExpressionMatrix</code>
     *                         with missing values from which the colored
     *                         table will be created
     * @param go the <code>GOFrontEnd</code>
     * @param parent the GUI's parent <code>Component</code>
     * @throws InvalidNodeObjectTypeException if <code>expressionMatrix</code>
     *                     is not of a valid object type
     * @see #TableColorsWithMissingValues(Object[][] rowData, Object[]
     *   columnNames)
     * @see TableValues#constructRowData(NodeObjectInterface nodeObject)
     * @see TableValues#constructColumnNames(NodeObjectInterface nodeObject)
     */
    public TableColorsWithMissingValues(AbstractExpressionMatrix expressionMatrix,
                                        GOFrontEnd go, Component parent) throws
            InvalidNodeObjectTypeException {
        this(constructRowData(expressionMatrix),
             constructColumnNames(expressionMatrix),
             expressionMatrix.getMissingValue());
        this.goFrontEnd = go;
        this.parent = parent;
    }

    /**
     * Creates a new colored table with data from the given expression matrix
     * with missing values.
     *
     * @param expressionMatrix the <code>DiscretizedExpressionMatrix</code>
     *                         with missing values from which the colored
     *                         table will be created
     * @param go the <code>GOFrontEnd</code>
     * @param parent the GUI's parent <code>Component</code>
     * @throws InvalidNodeObjectTypeException if <code>expressionMatrix</code>
     *                     is not of a valid object type
     * @see #TableColorsWithMissingValues(Object[][] rowData, Object[]
     *   columnNames)
     * @see TableValues#constructRowData(NodeObjectInterface nodeObject)
     * @see TableValues#constructColumnNames(NodeObjectInterface nodeObject)
     */
    public TableColorsWithMissingValues(DiscretizedExpressionMatrix expressionMatrix,
                                        GOFrontEnd go, Component parent) throws
            InvalidNodeObjectTypeException {
        this(constructRowData(expressionMatrix),
             constructColumnNames(expressionMatrix),
             expressionMatrix.getOriginalExpressionMatrix().getMissingValue());
        this.goFrontEnd = go;
        this.parent = parent;
    }

    /**
     * Creates a new colored table with data from the given expression matrix
     * with missing values.
     *
     * @param expressionMatrix the <code>ExpressionMatrix</code> with missing
     *                         values from which the colored table will be created
     * @throws InvalidNodeObjectTypeException if <code>expressionMatrix</code>
     *                     is not of a valid object type
     * @see #TableColorsWithMissingValues(Object[][] rowData, Object[]
     *   columnNames)
     * @see TableValues#constructRowData(NodeObjectInterface nodeObject)
     * @see TableValues#constructColumnNames(NodeObjectInterface nodeObject)
     */
    public TableColorsWithMissingValues(AbstractExpressionMatrix expressionMatrix)
            throws InvalidNodeObjectTypeException {
        this(constructRowData(expressionMatrix),
             constructColumnNames(expressionMatrix),
             expressionMatrix.getMissingValue());
    }

    /**
     * Creates a new colored table with data from the given expression matrix
     * with missing values.
     *
     * @param expressionMatrix the <code>DiscretizedExpressionMatrix</code>
     *                         with missing values from which the colored
     *                         table will be created
     * @throws InvalidNodeObjectTypeException if <code>expressionMatrix</code>
     *                     is not of a valid object type
     * @see #TableColorsWithMissingValues(Object[][] rowData, Object[]
     *   columnNames)
     * @see TableValues#constructRowData(NodeObjectInterface nodeObject)
     * @see TableValues#constructColumnNames(NodeObjectInterface nodeObject)
     */
    public TableColorsWithMissingValues(
            DiscretizedExpressionMatrix expressionMatrix)
            throws InvalidNodeObjectTypeException {
        this(constructRowData(expressionMatrix),
             constructColumnNames(expressionMatrix),
             expressionMatrix.getOriginalExpressionMatrix().getMissingValue());
    }

    /**
     * Creates a new colored table with data from the given bicluster
     * with missing values.
     *
     * @param bicluster the <code>Bicluster</code> from which the colored
     *                  table will be created
     * @param go the <code>GOFrontEnd</code>
     * @param parent the GUI's parent <code>Component</code>
     * @throws InvalidNodeObjectTypeException if <code>expressionMatrix</code>
     *                     is not of a valid object type
     * @see #TableColorsWithMissingValues(Object[][] rowData, Object[]
     *   columnNames)
     * @see TableValues#constructRowData(NodeObjectInterface nodeObject)
     * @see TableValues#constructColumnNames(NodeObjectInterface nodeObject)
     */
    public TableColorsWithMissingValues(Bicluster bicluster,
                                        GOFrontEnd go, Component parent)
            throws InvalidNodeObjectTypeException {
        this(constructRowData(bicluster),
             constructColumnNames(bicluster),
             bicluster.getMissingValue());
        this.goFrontEnd = go;
        this.parent = parent;
    }

    /**
     * Creates a new colored table with data from the given bicluster
     * with missing values.
     *
     * @param bicluster the <code>Bicluster</code> from which the colored
     *                  table will be created
     * @throws InvalidNodeObjectTypeException if <code>expressionMatrix</code>
     *                     is not of a valid object type
     * @see #TableColorsWithMissingValues(Object[][] rowData, Object[]
     *   columnNames)
     * @see TableValues#constructRowData(NodeObjectInterface nodeObject)
     * @see TableValues#constructColumnNames(NodeObjectInterface nodeObject)
     */
    public TableColorsWithMissingValues(Bicluster bicluster) throws
            InvalidNodeObjectTypeException {
        this(constructRowData(bicluster),
             constructColumnNames(bicluster),
             bicluster.getMissingValue());
    }

    /**
     * Creates a new colored table with the given row data and columns' names.
     *
     * @param rowData the <code>Object[][]</code> array with the data to fill
     *   the rows of the table
     * @param columnNames the <code>Object[][]</code> array with the data to
     *   fill the header of the table
     * @param missingValue the <code>float</code> value that marks missing values
     * @see #calcMeanDividerValues(Object[][] rowData)
     * @see TableColors#TableColors(Object[][] rowData, Object[] columnNames) /
     */
    private TableColorsWithMissingValues(Object[][] rowData,
                                         Object[] columnNames,
                                         float missingValue) {
        super(rowData, columnNames, false);
        this.missingValue = missingValue;
        this.calcMeanDividerValues(rowData);
    }

    /**
     * Overrides <code>TableColors</code>'s <code>prepareRenderer</code> method.
     * Prepares the renderer by querying the data model for the value and
     * selection state of the cell at <code>row</code>, <code>column</code>.
     * Returns the component (may be a <code>Component</code>
     * or a <code>JComponent</code>) under the event location.
     *
     * @param renderer  the <code>TableCellRenderer</code> to prepare
     * @param row       the row of the cell to render, where 0 is the first row
     * @param column    the column of the cell to render,
     *			where 0 is the first column
     * @return          the <code>Component</code> under the event location
     *
     * @see TableColors#prepareRenderer(TableCellRenderer renderer, int row, int column)
     */
    public Component prepareRenderer(TableCellRenderer renderer, int row,
                                     int column) {
        Component component = super.prepareRenderer(renderer, row, column);
        if (column != 0) {
            /**
             * this cell is not on the first column, i.e., contains a gene
             * expression value (not a gene name)
             */
            float value = ((Float)this.getValueAt(row, column)).floatValue();
            if (value == missingValue) {
                component.setBackground(HeatmapColorScheme.getMissingValueColor());
                component.setForeground(HeatmapColorScheme.getMissingValueColor());
            } else
            if (value == meanValue) {
                /**
                 * this cell's value equals the mean value, so it's painted black
                 */
                component.setBackground(HeatmapColorScheme.getMeanValueColor());
                component.setForeground(HeatmapColorScheme.getMeanValueColor());
            } else {
                /**
                 * this cell's value is different from the mean value; it'll
                 * be painted green or red; find the green or red color
                 * component <code>rgb</code>
                 */
                int rgb = Math.round(Math.abs(value - meanValue) / divider);

                /**
                 * limit the red/green color component, so that it is not
                 * higher than 255 nor lower than 0 (since rgb components'
                 * values always range between 0 and 255)
                 */
                if (rgb > 255) {
                    rgb = 255;
                } else
                if (rgb < 0) {
                    rgb = 0;
                }

                Color color;
                if (value > meanValue) {
                    /**
                     * the actual cell's value is higher than the mean value;
                     * so, the cell is colored red; red component value was
                     * calculated before:
                     * Red = <code>rgb</code>, Green = 0, Blue = 0
                     */
                    color = HeatmapColorScheme.getColorAboveMean(rgb);
                } else {
                    /**
                     * the actual cell's value is lower than the mean value;
                     * so, the cell is colored green; green component value was
                     * calculated before:
                     * Red = 0, Green = <code>rgb</code>, Blue = 0
                     */
                    color = HeatmapColorScheme.getColorBelowMean(rgb);
                }

                /**
                 * turn the prepared color brighter and set background and
                 * foreground colors with the prepared color
                 */
                for (int i = 0; i < 4; i++) {
                    color = color.brighter();
                }
                component.setBackground(color);
                component.setForeground(color);
            }
        } else
        /**
         * this cell is on the first column (it's one of those which
         * contain genes' names)
         */
        if (!isCellSelected(row, column)) {
            /**
             * this cell is on the first column and it is not selected by
             * the user; so, the background is set to default
             */
            component.setBackground(getBackground());
        }

        return component;
    }

    /**
     * Overrides <code>TableValues</code>'s <code>getToolTipText</code> method
     * in order to allow the renderer's tips to be used if they have text set.
     *
     * @param event MouseEvent the event that triggered the tool tip display
     * @return String the message set as tool tip for the pointed table cell
     * @see javax.swing.JTable#getToolTipText(MouseEvent event)
     */
    public String getToolTipText(MouseEvent event) {
        String tip = null;
        Point p = event.getPoint();
        int rowIndex = rowAtPoint(p);
        int colIndex = columnAtPoint(p);
        int realColumnIndex = convertColumnIndexToModel(colIndex);

        if (realColumnIndex == 0) {
            /**
             * cell is in the first column, so it contains the name of a gene
             * set the tex tip for this cell, something like: "Gene X"
             */
            tip = "Gene: " + getValueAt(rowIndex, 0);
        } else
        if ((rowIndex >= 0) && (rowIndex < this.getRowCount()) &&
            (realColumnIndex > 0) && (realColumnIndex < this.getColumnCount())) {
            /**
             * cell is in a column that is not the first one, so it contains
             * a float value (the gene expression value);
             * set the tip for this cell, something like:
             * "(gene_name, condition_name) = gene_expression_value"
             * if it's not a missing value, and:
             * "missing value"
             * it it is a missing value
             */
            if (((Float) getValueAt(rowIndex, realColumnIndex)).floatValue() ==
                missingValue) {
                tip = "missing value";
            } else {
                tip = "(" + getValueAt(rowIndex, 0) +
                      ", " + getColumnName(realColumnIndex) +
                      ") = " + getValueAt(rowIndex, realColumnIndex);
            }
        } else {
            tip = super.getToolTipText(event);
        }
        return tip;
    }

    /**
     * Calculates mean and divider values for the given row data, excluding
     * the value that marks cells with missing values. Mean and divider values
     * are used by <code>prepareRenderer</code> method to calculate the green
     * or red color component to paint the cell.
     * <p>
     * Mean value is calculated like this:<br/>
     * <b>meanValue = (minValue + maxValue) / 2</b><br/>
     * where minValue is the minimum value of all the expression matrix's values
     * and maxValue is the maximum value of all.
     * </p>
     * <p>
     * Divider value is calculated like this:<br/>
     * <b>divider = (meanValue-minValue) / 255</b><br/>
     * where minValue, maxValue and meanValue are the same mentioned above.
     * </p>
     *
     * @param rowData the <code>Object[][]</code> row data that includes all
     *                gene expression values, from which the mean and divider
     *                values are calculated
     */
    private void calcMeanDividerValues(Object[][] rowData) {
        float minValue;
        float maxValue;
        ;
        float value;
        int i = 0, j = 0;

        /**
         * find the first value that is not a missing value
         */
        for (i = 0; i < rowData.length; i++) {
            for (j = 1; j < rowData[0].length; i++) {
                if (((Float) (rowData[i][j])).floatValue() != missingValue) {
                    break;
                }
            }
            if (j != rowData[0].length) {
                break;
            }
        }
        if ((i == rowData.length) && (j == rowData[0].length)) {
            return;
        }
        minValue = ((Float) rowData[i][j]).floatValue();
        maxValue = minValue;

        /**
         * find minimum and maximum values of the row data values
         */
        for (Object[] row : rowData) {
            for (i = 1; i < row.length; i++) {
                value = ((Float) row[i]).floatValue();
                /**
                 * if value marks a missing value, ignore it;
                 * otherwise, check if it is a local maximum or minimum
                 */
                if (value != missingValue) {
                    if (value < minValue) {
                        minValue = value;
                    }
                    if (value > maxValue) {
                        maxValue = value;
                    }
                }
            }
        }

        /**
         * calculate mean and divider values
         */
        meanValue = (minValue + maxValue) / 2;
        divider = (meanValue - minValue) / 255;
    }

    /**
     * Returns the value used to mark cells which correspond to missing
     * values on the original gene expression data.
     *
     * @return the <code>float</code> that marks missing values
     */
    public float getMissingValue() {
        return missingValue;
    }
}
