package biggests.tables;

import java.awt.*;
import java.awt.event.*;
import java.text.*;
import java.util.*;

import javax.swing.*;
import javax.swing.table.*;

import biggests.utils.*;
import smadeira.biclustering.*;
import smadeira.utils.BiologicalRelevanceSummary;

/**
 * <p>Title: Table of Functions for a Group of Biclusters</p>
 *
 * <p>Description: Defines a table for displaying data of function
 *                 analysis performed on a given a group of biclusters.<br/>
 *                 It contains the following information:<br/>
 *                 1st column - Bicluster ID<br/>
 *                 2nd column - Number of Genes<br/>
 *                 3rd column - Number of Conditions<br/>
 *                 4th column - Best p-value<br/>
 *                 5th column - Best corrected p-value<br/>
 *                 6th column - Number of significant functions<br/>
 *                 7th column - Number of highly significant functions<br/>
 *                 8th column - P-value user's threshold<br/>
 *                 9th column - Number of significant functions (user's threshold)<br/>
 * Note: if the bicluster group is a post-processed one, an an extra
 *       column is added at the beggining, holding the post-processed
 *       bicluster id.
 * </p>
 *
 * <p>Copyright:   Copyright (C) 2007  Joana Gonçalves
 *                 This program is free software; you can redistribute
 *                 it and/or modify it under the terms of the GNU General
 *                 Public License as published by the Free Software
 *                 Foundation; either version 3 of the License, or
 *                 any later version.
 *
 *                 This program is distributed in the hope that it will
 *                 be useful, but WITHOUT ANY WARRANTY; without even the
 *                 implied warranty of MERCHANTABILITY or FITNESS FOR A
 *                 PARTICULAR PURPOSE.  See the
 *                 <a href="http://www.gnu.org/licenses/gpl.html">
 *                 GNU General Public License</a> for more details.
 * </p>
 *
 * @author Joana P. Gonçalves
 */
/* ------------------------------------------------------------------------------------------------ ...<br/>
 * | Bicluster ID | #Genes | #Conditions | Best p-value | Best corr. p-value | #Signif. functions | ...<br/>
 * |----------------------------------------------------------------------------------------------| ...<br/>
 * |      1       |  136   |      6      | 0.0357434694 |   0.00082635340    |        3           | ...<br/>
 * |      2       |  862   |     17      | 0.7639486945 |         1.0        |        0           | ...<br/>
 * | ............ | ...... | ........... | ............ | .................. | .................. | ...<br/>
 * | ............ | ...... | ........... | ............ | .................. | .................. | ...<br/>
 * ------------------------------------------------------------------------------------------------ ...<br/>
 */
public class TableFunctionGroup extends JTable {
    /**
     * Default color to mark rows that contain biclusters with significant terms.
     */
    public static final Color SIGNIFICANT_COLOR = new Color(89, 255, 89);

    /**
     * Default color to user in even rows (when they are not selected).
     */
    public static final Color EVEN_ROWS_COLOR = new Color(228, 238, 250);

    /**
     * Default color to user in odd rows (when they are not selected).
     */
    public static final Color ODD_ROWS_COLOR = Color.WHITE;

    /**
     * States if this function table is for a post-processed bicluster group or not.
     */
    private boolean postProcessed;

    /**
     * Color to mark rows that contain biclusters with significant terms.
     */
    private Color significantColor;

    /**
     * Color to mark even rows (when they are not selected).
     */
    private Color evenRowsColor;

    /**
     * Color to mark odd rows (when they are not selected).
     */
    private Color oddRowsColor;

    /**
     * Creates a new function table from the data of a biclusters group object.
     *
     * @param biclustering the <code>Biclustering</code> from which the
     *                   function table is created
     * @param biologicalRelevanceSummary the <code>Hashtable</code> which
     *        contains the biological relevance summaries for all
     *        biclusters' group biclusters
     * @param postProcessed <code>boolean</code> <code>true</code> if
     *        <code>biclustering</code> is a post-processed biclusters
     *        group object
     * @throws InvalidNodeObjectTypeException when the type of the biclustering
     *                   object is not valid
     * @see #constructRowData(Biclustering biclustering)
     * @see #TableFunctionGroup(Object[][] rowData, Object[] columnNames))
     */
    public TableFunctionGroup(Biclustering biclustering,
            Hashtable<Integer,BiologicalRelevanceSummary>
            biologicalRelevanceSummary,
            boolean postProcessed) throws
            InvalidNodeObjectTypeException {
        this(constructRowData(biclustering, biologicalRelevanceSummary,
                              postProcessed),
             columnNames(postProcessed), postProcessed);
        this.postProcessed = postProcessed;
    }

    /**
     * Creates a new function table from the data of a biclusters group object.
     *
     * @param rowData the <code>Object[][]</code> row data (for all table
     *                cells except the ones from the table header)
     * @param columnNames the <code>Object[]</code> column names (for
     *                the table header)
     * @param postProcessed <code>boolean</code> <code>true</code> if
     *        <code>biclustering</code> is a post-processed biclusters
     *        group object
     * @see #TableFunctionGroup(TableSorterFunctionGroup tableSorterModel)
     * @see applicationgui.utils.TableSorter#TableSorter(TableModel tableModel)
     * @see javax.swing.table.DefaultTableModel#DefaultTableModel(Object[][] rowData, Object[] columnNames)
     */
    protected TableFunctionGroup(Object[][] rowData, Object[] columnNames,
                                 boolean postProcessed) {
        /**
         * create a new JTable with a table sorter model, which incorporates
         * the specified row data and column names
         */
        this(new TableSorterFunctionGroup(new DefaultTableModel(rowData,
                columnNames), postProcessed));
    }

    /**
     * Creates a new function table from the data of a biclusters group object.
     *
     * @param tableSorterModel the <code>TableSorter</code> model for the table
     * @see javax.swing.JTable#JTable(TableModel dm)
     */
    protected TableFunctionGroup(TableSorterFunctionGroup tableSorterModel) {
        /**
         * create a new JTable object with a table sorter model derived from the
         * given table model
         */
        super(tableSorterModel);

        /**
         * set table header of this TableValues as the table header of the
         * created table sorter model
         */
        ((TableSorterFunctionGroup)this.getModel()).setTableHeader(this.
                getTableHeader());

        this.getTableHeader().setReorderingAllowed(false);

        this.setDefaultRenderer(Float.class, new FloatCellRenderer());

        /**
         * set preferred size for all table's cells
         */
        Enumeration<TableColumn> e = this.getColumnModel().getColumns();
        while (e.hasMoreElements()) {
            e.nextElement().setPreferredWidth(150);
        }

        /**
         * set header font to bold style
         */
        JTableHeader tableHeader = this.getTableHeader();
        Font headerFont = tableHeader.getFont();
        tableHeader.setFont(new Font(headerFont.getName(), Font.BOLD,
                                     headerFont.getSize()));

        this.significantColor = this.SIGNIFICANT_COLOR;
        this.evenRowsColor = this.EVEN_ROWS_COLOR;
        this.oddRowsColor = this.ODD_ROWS_COLOR;
    }

    /**
     * Overrides <code>JTable</code>'s <code>prepareRenderer</code> method.
     * Prepares the renderer by querying the data model for the
     * value and selection state
     * of the cell at <code>row</code>, <code>column</code>.
     * Returns the component (may be a <code>Component</code>
     * or a <code>JComponent</code>) under the event location.
     *
     * @param renderer  the <code>TableCellRenderer</code> to prepare
     * @param row       the row of the cell to render, where 0 is the first row
     * @param column    the column of the cell to render,
     *			where 0 is the first column
     * @return          the <code>Component</code> under the event location
     * @see javax.swing.JTable#prepareRenderer(TableCellRenderer renderer, int row, int column)
     */
    public Component prepareRenderer(TableCellRenderer renderer, int row,
                                     int column) {
        /**
         * get the component prepared from JTable's prepareRenderer method
         */
        Component c = super.prepareRenderer(renderer, row, column);
        if (postProcessed) {
            if ((row % 2 == 0) && (!isCellSelected(row, column))) {
                /**
                 * cell is in row numbered even and is not selected
                 */
                int value = ((Integer)this.getValueAt(row, 9)).intValue();
                if (value > 0) {
                    /**
                     * signal a row which corresponds to a post-processed
                     * bicluster with significant functions
                     */
                    c.setBackground(this.significantColor);
                } else {
                    /**
                     * set the color of the cell to light blue
                     * (the color of even rows)
                     */
                    c.setBackground(this.evenRowsColor);
                }
            } else
            if (!isCellSelected(row, column)) {
                int value = ((Integer)this.getValueAt(row, 9)).intValue();
                if (value > 0) {
                    /**
                     * signal a row which corresponds to a post-processed
                     * bicluster with significant functions
                     */
                    c.setBackground(this.significantColor);
                } else {
                    /**
                     * cell is in a row numbered odd and is not selected
                     * set the color of the cell to default background
                     */
                    c.setBackground(this.oddRowsColor);
                }
            }
            return c;
        }

        if ((row % 2 == 0) && (!isCellSelected(row, column))) {
            /**
             * cell is in row numbered even and is not selected
             */
            int value = ((Integer)this.getValueAt(row, 8)).intValue();
            if (value > 0) {
                /**
                 * signal a row which corresponds to a bicluster with
                 * significant functions
                 */
                c.setBackground(this.significantColor);
            } else {
                /**
                 * set the color of the cell to light blue
                 * (the color of even rows)
                 */
                c.setBackground(this.evenRowsColor);
            }
        } else
        if (!isCellSelected(row, column)) {
            int value = ((Integer)this.getValueAt(row, 8)).intValue();
            if (value > 0) {
                /**
                 * signal a row which corresponds to a bicluster with
                 * significant functions
                 */
                c.setBackground(this.significantColor);
            } else {
                /**
                 * cell is in a row numbered odd and is not selected
                 * set the color of the cell to default background
                 */
                c.setBackground(this.oddRowsColor);
            }
        }
        return c;
    }

    /**
     * Overrides <code>JTable</code>'s <code>getToolTipText</code> method
     * in order to allow the renderer's tips to be used if they have text set.
     *
     * @param event MouseEvent the event that triggered the tool tip display
     * @return String the message set as tool tip for the pointed table cell
     * @see javax.swing.JTable#getToolTipText(MouseEvent event)
     */
    public String getToolTipText(MouseEvent event) {
        String tip = null;
        Point p = event.getPoint();
        int rowIndex = rowAtPoint(p);
        int colIndex = columnAtPoint(p);
        int realColumnIndex = convertColumnIndexToModel(colIndex);

        if (rowIndex < 0) {
            return "";
        }

        if (realColumnIndex == 0) {
            /**
             * cell is in the first column, so it contains
             * a bicluster or post-processed bicluster id;
             * set the text tip for this cell, something like:
             * "Bicluster ID 14" or "Post-Processed ID 14"
             */
            tip = completeColumnNames(postProcessed)[0] + " " +
                  getValueAt(rowIndex, 0);
        } else
        if ((realColumnIndex == 1) && postProcessed) {
            tip = completeColumnNames(postProcessed)[1] + " " +
                  getValueAt(rowIndex, 1);
        } else
        if ((rowIndex >= 0) && (rowIndex < this.getRowCount()) &&
            (realColumnIndex > 0) && (realColumnIndex < this.getColumnCount())) {
            /**
             * cell is in a column that is not the first one;
             * set the tip for this cell, something like:
             * "(Bicluster 14, Number of Significant Functions) = 3"
             */
            tip = "(" + completeColumnNames(postProcessed)[0] + " " +
                  getValueAt(rowIndex, 0) +
                  ", " + completeColumnNames(postProcessed)[realColumnIndex] +
                  ") = " + getValueAt(rowIndex, realColumnIndex);
        } else {
            tip = super.getToolTipText(event);
        }
        return tip;
    }

    /**
     * Overrides <code>JTable</code>'s <code>createDefaultTableHeader</code>
     * method.<br/>
     * Creates the default table header for a values' table, that is, a header
     * with specific text tool tips for this kind of table.
     *
     * @return an instance of <code>TableHeaderFunction</code>, i.e, a specific table
     *         header for a table of functions
     * @see javax.swing.JTable#columnModel
     * @see javax.swing.JTable#createDefaultTableHeader()
     * @see applicationgui.utils.TableHeader#TableHeader(TableColumnModel cm)
     */
    protected JTableHeader createDefaultTableHeader() {
        return new TableHeaderFunction(columnModel);
    }

    /**
     * Constructs the row data matrix with data retrieved from the biclustering
     * object.<br/>
     * <p>
     * The final structure of the row data is:<br/>
     * for GROUPS OF BICLUSTERS<br/>
     * 1st column - Bicluster ID<br/>
     * 2nd column - Number of Genes<br/>
     * 3rd column - Number of Conditions<br/>
     * 4th column - Best p-value<br/>
     * 5th column - Best corrected p-value<br/>
     * 6th column - Number of significant functions<br/>
     * 7th column - Number of highly significant functions<br/>
     * 8th column - P-value user's threshold
     * 9th column - Number of significant functions (user's threshold)
     *  <br/>
     * for POST-PROCESSED GROUPS OF BICLUSTERS<br/>
     * 1st column - Post-Processed ID<br/>
     * 2nd column - Bicluster ID<br/>
     * 3rd column - Number of Genes<br/>
     * 4th column - Number of Conditions<br/>
     * 5th column - Best p-value<br/>
     * 6th column - Best corrected p-value<br/>
     * 7th column - Number of significant functions<br/>
     * 8th column - Number of highly significant functions<br/>
     * 9th column - P-value user's threshold
     * 10th column - Number of significant functions (user's threshold)
     * </p>
     *
     * @param biclustering the <code>Biclustering</code> object which provides the
     *                   data needed to create the row data matrix
     * @param biologicalRelevanceSummary <code>Hashtable</code> contains
     *        the summary of function analysis results for all the
     *        biclusters in the group of biclusters (<code>biclustering</code>)
     * @param postProcessed <code>boolean</code> <code>true</code> if
     *        <code>biclustering</code> is a post-processed biclusters
     *        group object
     * @return the <code>Object[][]</code> row data matrix, ready to be used by
     *         <code>TableFunctionGroup</code>'s constructor
     * @throws InvalidNodeObjectTypeException when the type of the biclustering object is not valid
     */
    protected static Object[][] constructRowData(
            Biclustering biclustering,
            Hashtable<Integer,BiologicalRelevanceSummary> biologicalRelevanceSummary,
            boolean postProcessed) throws
            InvalidNodeObjectTypeException {
        Object[][] rowData = null;
        Bicluster[] biclusters = biclustering.getSetOfBiclusters();
        if (biclusters == null || biclusters.length == 0) {
            return null;
        }

        if (postProcessed) {
            rowData = new Object[biclusters.length][10];
            int i = 0;
            for (Bicluster bicluster : biclusters) {
                BiologicalRelevanceSummary summary =
                        biologicalRelevanceSummary.get(bicluster.getID());
                rowData[i][0] = Integer.valueOf(i + 1);
                rowData[i][1] = Integer.valueOf(bicluster.getID());
                rowData[i][2] = Integer.valueOf(bicluster.getNumberOfGenes());
                rowData[i][3] = Integer.valueOf(bicluster.getNumberOfConditions());
                rowData[i][4] = Float.valueOf((float) summary.getBestPvalue());
                rowData[i][5] = Float.valueOf((float) summary.getBestCorrectedPvalue());
                rowData[i][6] = Float.valueOf((float) summary.getNumberOfSignificantCorrectedTerms());
                rowData[i][7] = Float.valueOf((float) summary.getNumberOfHighlySignificantCorrectedTerms());
                rowData[i][8] = Float.valueOf((float) summary.getSignificanceUserThreshold());
                rowData[i][9] = Integer.valueOf(summary.getNumberOfUserThresholdSignificantCorrectedTerms());
                i++;
            }
            return rowData;
        }

        rowData = new Object[biclusters.length][9];
        int i = 0;
        for (Bicluster bicluster : biclusters) {
            BiologicalRelevanceSummary summary =
                    biologicalRelevanceSummary.get(bicluster.getID());
            rowData[i][0] = Integer.valueOf(bicluster.getID());
            rowData[i][1] = Integer.valueOf(bicluster.getNumberOfGenes());
            rowData[i][2] = Integer.valueOf(bicluster.getNumberOfConditions());
            rowData[i][3] = Float.valueOf((float) summary.getBestPvalue());
            rowData[i][4] = Float.valueOf((float) summary.getBestCorrectedPvalue());
            rowData[i][5] = Float.valueOf((float) summary.getNumberOfSignificantCorrectedTerms());
            rowData[i][6] = Float.valueOf((float) summary.getNumberOfHighlySignificantCorrectedTerms());
            rowData[i][7] = Float.valueOf((float) summary.getSignificanceUserThreshold());
            rowData[i][8] = Integer.valueOf(summary.getNumberOfUserThresholdSignificantCorrectedTerms());
            i++;
        }
        return rowData;
    }

    /**
     * Returns the abbreviated names for the columns of this functions table.
     *
     * The abbreviated column names for the header of this table of functions.<br/>
     * GROUP OF BICLUSTERS:<br/>
     * 1st column - Bicluster ID<br/>
     * 2nd column - # Genes<br/>
     * 3rd column - # Conditions<br/>
     * 4th column - Best p-value<br/>
     * 5th column - Best corrected p-value<br/>
     * 6th column - # Significant terms<br/>
     * 7th column - # Highly sig terms<br/>
     * 8th column - Significance threshold<br/>
     * 9th column - # Sig terms (threshold)<br/>
     *
     * POST-PROCESSED GROUP OF BICLUSTERS:<br/>
     * 1st column - Post-Processed ID<br/>
     * 2nd column - Parent ID<br/>
     * 3rd column - # Genes<br/>
     * 4th column - # Conditions<br/>
     * 5th column - Best p-value<br/>
     * 6th column - Best corrected p-value<br/>
     * 7th column - # Significant terms<br/>
     * 8th column - # Highly sig terms<br/>
     * 9th column - Significance threshold<br/>
     * 10th column - # Sig terms (threshold)<br/>
     *
     * @param postProcessed <code>boolean</code> <code>true</code> if
     *        <code>biclustering</code> is a post-processed biclusters
     *        group object
     * @return the <code>String[]</code> set of abbreviated columns names
     */
    public static String[] columnNames(boolean postProcessed) {
        if (postProcessed) {
            String[] cNames = {"Post-Processed ID",
                              "Parent ID",
                              "# Genes",
                              "# Time points",
                              "Best p-value",
                              "Best corrected p-value",
                              "# Significant terms",
                              "# Highly sig terms",
                              "Significance threshold",
                              "# Sig terms (threshold)"};
            return cNames;
        }
        String[] cNames = {"Bicluster ID",
                          "# Genes",
                          "# Time points",
                          "Best p-value",
                          "Best corrected p-value",
                          "# Significant terms",
                          "# Highly sig terms",
                          "Significance threshold",
                          "# Sig terms (threshold)"};
        return cNames;
    }

    /**
     * Returns the complete names of the columns of this functions table.
     *
     * Complete column names for the header of this table of functions:<br/>
     * BICLUSTERS GROUP:<br/>
     * 1st column - Bicluster ID<br/>
     * 2nd column - Number of Genes<br/>
     * 3rd column - Number of Conditions<br/>
     * 4th column - Best p-value<br/>
     * 5th column - Best corrected p-value<br/>
     * 6th column - Number of Significant Rerms<br/>
     * 7th column - Number of Highly Significant Rerms<br/>
     * 8th column - Significance threshold<br/>
     * 9th column - Number of Significant Terms (below significance threshold)<br/>
     *
     * POST-PROCESSED BICLUSTER GROUP:<br/>
     * 1st column - Post-Processed ID<br/>
     * 2nd column - Bicluster ID<br/>
     * 3rd column - Number of Genes<br/>
     * 4th column - Number of Conditions<br/>
     * 5th column - Best p-value<br/>
     * 6th column - Best corrected p-value<br/>
     * 7th column - Number of Significant Terms<br/>
     * 8th column - Number of Highly Significant Terms<br/>
     * 9th column - Significance threshold<br/>
     * 10th column - Number of Significant Terms (below significance threshold)<br/>
     *
     * @param postProcessed <code>boolean</code> <code>true</code> if
     *        <code>biclustering</code> is a post-processed biclusters
     *        group object
     * @return the <code>String[]</code> set of complete columns names
     */
    public static String[] completeColumnNames(boolean postProcessed) {
        if (postProcessed) {
            String[] ccNames = {"Post-Processed ID",
                               "Parent ID",
                               "Number of genes",
                               "Number of time points",
                               "Best p-value",
                               "Best corrected p-value",
                               "Number of significant terms",
                               "Number of highly significant terms",
                               "Significance threshold",
                               "Number of significant terms (below significance threshold)"};
            return ccNames;
        }

        String[] ccNames = {"Bicluster ID",
                           "Number of genes",
                           "Number of time points",
                           "Best p-value",
                           "Best corrected p-value",
                           "Number of significant terms",
                           "Number of highly significant terms",
                           "Significance threshold",
                           "Number of significant terms (below significance threshold)"};
        return ccNames;
    }

    /**
     * Returns the <code>Color</code> used to mark significant terms.
     *
     * @return Color used to mark significant terms
     */
    public Color getSignificantColor() {
        return significantColor;
    }

    /**
     * Returns the <code>Color</code> used to mark even rows that are not selected.
     *
     * @return Color used to mark even rows that are not selected
     */
    public Color getEvenRowsColor() {
        return evenRowsColor;
    }

    /**
     * Returns the <code>Color</code> used to mark odd rows that are not selected.
     *
     * @return Color used to mark odd rows that are not selected
     */
    public Color getOddRowsColor() {
        return oddRowsColor;
    }

    /**
     * Sets a new color to mark significant terms.
     *
     * @param newSignificantColor <code>Color</code> to mark
     * significant terms
     */
    public void setSignificantColor(Color newSignificantColor) {
        significantColor = newSignificantColor;
    }

    /**
     * Sets a new color to mark even rows that are not selected.
     *
     * @param newEvenRowsColor <code>Color</code> to mark even
     * rows that are not selected
     */
    public void setEvenRowsColor(Color newEvenRowsColor) {
        evenRowsColor = newEvenRowsColor;
    }

    /**
     * Sets a new color to mark odd rows that are not selected.
     *
     * @param newOddRowsColor <code>Color</code> to mark odd
     * rows that are not selected
     */
    public void setOddRowsColor(Color newOddRowsColor) {
        oddRowsColor = newOddRowsColor;
    }

    /**
     * <p>Title: <code>NumberCellRenderer</code>: A renderer for cells which
     *           contain numbers.</p>
     */
    protected static class NumberCellRenderer extends DefaultTableCellRenderer.
            UIResource {
        /**
         * Constructs a new <code>NumberCellRenderer</code> for a cell.
         *
         * @see javax.swing.table.DefaultTableCellRenderer#DefaultTableCellRenderer()
         */
        public NumberCellRenderer() {
            super();
            setHorizontalAlignment(JLabel.RIGHT);
        }
    }


    /**
     * <p>Title: <code>FloatCellRenderer</code>: A renderer for cells which
     *           contain float numbers.</p>
     */
    protected static class FloatCellRenderer extends NumberCellRenderer {
        /**
         * The formatter which is used to format the float value.
         */
        NumberFormat formatter;

        /**
         * Constructs a new <code>FloatCellRenderer</code>.
         */
        public FloatCellRenderer() {
            super();
        }

        /**
         * Overrides <code>DefaultTableCellRenderer</code>'s
         * <code>setValue</code> method. Formats the value of the cell to display.
         *
         * @param value the <code>Object</code> containing the float value
         */
        public void setValue(Object value) {
            if (formatter == null) {
                formatter = NumberFormat.getInstance(Locale.ENGLISH);
                formatter.setMaximumFractionDigits(8);
                formatter.setMinimumFractionDigits(8);
            }
            setText((value == null) ? "" : formatter.format(value));
        }
    }
}
